# SPDX-FileCopyrightText: 2020-2025 Ivan Perevala <ivan95perevala@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Generator

import bpy
from bl_ui import space_statusbar
from bpy.types import Context, PropertyGroup, STATUSBAR_HT_header, UILayout, WindowManager
from bpy.props import StringProperty, EnumProperty, BoolProperty, IntProperty, CollectionProperty

from . _unique_name import eval_unique_name

__all__ = (
    "progress",
)


class progress:
    """
    Displays progressbar in statusbar.

    :cvar int PROGRESS_BAR_UI_UNITS: UI units [4...12] for each progressbar (label and icon size does not count). 6 by default (readonly).
    """

    _is_drawn = False
    _attrname = ""

    @staticmethod
    def _update_statusbar():
        if bpy.context.workspace:
            bpy.context.workspace.status_text_set(text=None)

    class ProgressPropertyItem(PropertyGroup):
        "Single progressbar indicator."

        identifier: StringProperty(
            maxlen=64,
            options={'HIDDEN'},
        )
        "Progressbar identifier name."

        def _common_value_update(self, context: Context) -> None:
            progress._update_statusbar()

        subtype: EnumProperty(
            items=(
                ('PERCENTAGE', "Percentage", ""),
                ('STEP', "Step", ""),
                ('CONTINUOUS', "Continuous", "")
            ),
            default='PERCENTAGE',
            options={'HIDDEN'},
            update=_common_value_update,  # type: ignore
        )
        "Progressbar subtype."

        valid: BoolProperty(
            default=True,
            update=_common_value_update,  # type: ignore
        )
        "If progressbar is still valid."

        label: StringProperty(
            default="Progress",
            options={'HIDDEN'},
            update=_common_value_update,  # type: ignore
        )
        "Progressbar label."

        num_steps: IntProperty(
            min=1,
            default=1,
            subtype='UNSIGNED',
            options={'HIDDEN'},
            update=_common_value_update,  # type: ignore
        )
        "Number of steps."

        step: IntProperty(
            min=0,
            default=0,
            subtype='UNSIGNED',
            options={'HIDDEN'},
            update=_common_value_update,  # type: ignore
        )
        "Current step."

    def _func_draw_progress(self, context: Context):
        layout: UILayout = self.layout  # type: ignore

        if hasattr(WindowManager, progress._attrname):
            for item in progress.valid_progress_items():
                row = layout.row(align=True)

                factor = item.step / item.num_steps

                match item.subtype:
                    case 'PERCENTAGE':
                        row.label(text=item.label)
                        row.progress(type='BAR', text=f"{factor * 100} %", factor=factor)
                    case 'STEP':
                        row.label(text=item.label)
                        row.progress(type='BAR', text=f"{item.step} / {item.num_steps}", factor=factor)
                    case 'CONTINUOUS':
                        row.progress(type='RING', text=item.label, factor=factor)
                        item.num_steps = 25
                        item.step += 1
                        if item.step >= 25:
                            item.step = 0

    @classmethod
    def progress_items(cls) -> tuple[ProgressPropertyItem]:
        return tuple(getattr(bpy.context.window_manager, cls._attrname, tuple()))

    @classmethod
    def valid_progress_items(cls) -> Generator[ProgressPropertyItem]:
        """
        Unfinished progressbar items generator.

        :yield: Unfinished progressbar.
        :rtype: Generator[:class:`ProgressPropertyItem`]
        """

        return (_ for _ in cls.progress_items() if _.valid)

    @classmethod
    def _get(cls, *, identifier: str) -> None | ProgressPropertyItem:
        for item in cls.progress_items():
            if item.identifier == identifier:
                return item

    @classmethod
    def get(cls, *, identifier: str = "") -> ProgressPropertyItem:
        item = cls._get(identifier=identifier)
        if item is None:
            if not cls._is_drawn:
                bpy.utils.register_class(progress.ProgressPropertyItem)
                cls._attrname = eval_unique_name(arr=WindowManager, prefix="bhq_", suffix="_progress")

                setattr(
                    WindowManager,
                    cls._attrname,
                    CollectionProperty(type=progress.ProgressPropertyItem, options={'HIDDEN'})
                )
                STATUSBAR_HT_header.append(cls._func_draw_progress)
                cls._update_statusbar()

            cls._is_drawn = True
            ret: progress.ProgressPropertyItem = getattr(bpy.context.window_manager, cls._attrname).add()
            ret.identifier = identifier
            return ret
        else:
            ret = item
            ret.valid = True
            return ret

    @classmethod
    def complete(cls, *, identifier: str):
        item = cls._get(identifier=identifier)
        if item:
            item.valid = False

            for _ in cls.valid_progress_items():
                return
            cls.release_all()

    @classmethod
    def release_all(cls):
        """
        Removes all progressbars and restores statusbar to original state.
        """

        if not cls._is_drawn:
            return

        assert (cls._attrname)
        delattr(WindowManager, cls._attrname)
        bpy.utils.unregister_class(progress.ProgressPropertyItem)

        STATUSBAR_HT_header.remove(cls._func_draw_progress)
        cls._update_statusbar()

        cls._is_drawn = False



# SPDX-FileCopyrightText: 2020-2024 Ivan Perevala <ivan95perevala@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import os

import bpy
import bpy.utils.previews
from bpy.types import ImagePreview

__all__ = (
    "IconsCache",
)


class IconsCache:
    """
    Abstract icons cache class.
    """

    __slots__ = (
        "_directory",
        "_data_identifiers",
        "_image_names",
        "_cache",
        "_pcoll_cache",
    )

    _directory: str
    "Icons source directory."

    _data_identifiers: tuple[str, ...]
    "Icons data identifiers."

    _image_names: tuple[str, ...]
    "Icons image identifiers."

    _cache: dict[str, int]
    "Icons map: `identifier: icon_value`"

    _pcoll_cache: None | bpy.utils.previews.ImagePreviewCollection
    "Image icons cache"

    def _intern_initialize_from_data_files(self):
        for identifier in self._data_identifiers:
            try:
                icon_value = bpy.app.icons.new_triangles_from_file(os.path.join(self._directory, f"{identifier}.dat"))
            except ValueError:
                # log.warning(f"Unable to load icon \"{identifier}\"")
                icon_value = 0

            self._cache[identifier] = icon_value

    def _intern_initialize_from_image_files(self):
        if self._image_names:
            pcoll = bpy.utils.previews.new()
            for name in self._image_names:
                prv: ImagePreview = pcoll.load(name, os.path.join(self._directory, name), 'IMAGE')
                self._cache[os.path.splitext(name)[0]] = prv.icon_id
            self._pcoll_cache = pcoll

    def __init__(self, *, directory: str = "", data_identifiers: tuple[str, ...] = (), image_names: tuple[str, ...] = ()):
        self._directory = directory
        self._data_identifiers = data_identifiers
        self._image_names = image_names
        self._cache = dict()
        self._pcoll_cache = None

        self._intern_initialize_from_data_files()
        self._intern_initialize_from_image_files()

    def release(self):
        if self._pcoll_cache is not None:
            bpy.utils.previews.remove(self._pcoll_cache)
            self._pcoll_cache = None

        self._cache.clear()

    def get_id(self, identifier: str) -> int:
        return self._cache.get(identifier, 0)



# SPDX-FileCopyrightText: 2020-2024 Ivan Perevala <ivan95perevala@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import os

import bpy
from bpy.types import UILayout, Menu

__all__ = (
    "copy_default_presets_from",
    "template_preset",
)


def copy_default_presets_from(*, src_root: str):
    """
    Copies preset files from extension directory to Blender preset directory.

    :param src_root: Preset directory.
    :type src_root: str
    """

    for root, _dir, files in os.walk(src_root):
        for filename in files:
            rel_dir = os.path.relpath(root, src_root)
            src_fp = os.path.join(root, filename)

            tar_dir = bpy.utils.user_resource('SCRIPTS', path=os.path.join("presets", rel_dir), create=True)
            if not tar_dir:
                print("Failed to create presets path")
                return

            tar_fp = os.path.join(tar_dir, filename)

            with open(src_fp, 'r', encoding="utf-8") as src_file, open(tar_fp, 'w', encoding="utf-8") as tar_file:
                tar_file.write(src_file.read())


def template_preset(layout: UILayout, *, menu: Menu, operator: str) -> None:
    """
    Template for displaying presets in the UI.

    :param layout: Current UI layout.
    :type layout: `UILayout`_
    :param menu: Menu class to be used for displaying list of presets.
    :type menu: 'Menu'_
    :param operator: Operator's ``bl_idname`` identifier to be used for adding and removing presets.
    :type operator: str
    """

    row = layout.row(align=True)
    row.use_property_split = False

    row.menu(menu=menu.__name__, text=menu.bl_label)
    row.operator(operator=operator, text="", icon='ADD')
    row.operator(operator=operator, text="", icon='REMOVE').remove_active = True



# SPDX-FileCopyrightText: 2020-2025 Ivan Perevala <ivan95perevala@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import random
import string
from typing import Iterable

import bpy


__all__ = (
    "eval_unique_name",
)


def eval_unique_name(*, arr: Iterable, prefix: str = "", suffix: str = "") -> str:
    """
    Evaluates random unique name for new array items with pre-defined prefix and suffix. Might be used for ``bpy.data``
    and ``bpy.ops`` to register temporary data blocks.

    :param arr: Existing array.
    :type arr: Iterable
    :param prefix: Name prefix. In case of `bpy.ops` as an array prefix:
        ``bpy.ops.[prefix][unique name][suffix]`` - this might be used as temporary operator's `bl_idname` field.
    :type prefix: str, optional
    :param suffix: Name suffix, default to "".
    :type suffix: str, optional
    :return: Unique name.
    :rtype: str
    """

    if arr is bpy.ops:
        ret = prefix + '.' + str().join(random.sample(string.ascii_lowercase, k=10)) + suffix
        if isinstance(getattr(getattr(arr, ret, None), "bl_idname", None), str):
            return eval_unique_name(arr=arr, prefix=prefix, suffix=suffix)
        return ret
    else:
        ret = prefix + str().join(random.sample(string.ascii_letters, k=5)) + suffix
        if hasattr(arr, ret) or (isinstance(arr, Iterable) and ret in arr):
            return eval_unique_name(arr=arr, prefix=prefix, suffix=suffix)
        return ret



# SPDX-FileCopyrightText: 2020-2024 Ivan Perevala <ivan95perevala@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
This module is a collection of utility functionality for UI elements in the Blender extensions.
It includes functions for unique naming, text wrapping, developer extras, progress handling,
icon caching, and preset management.
"""

from __future__ import annotations

if __debug__:
    def __reload_submodules(lc):
        import importlib

        if "_unique_name" in lc:
            importlib.reload(_unique_name)
        if "_wrapped_text" in lc:
            importlib.reload(_wrapped_text)
        if "_developer_extras" in lc:
            importlib.reload(_developer_extras)
        if "_progress" in lc:
            importlib.reload(_progress)
        if "_icons_cache" in lc:
            importlib.reload(_icons_cache)
        if "_preset" in lc:
            importlib.reload(_preset)

    __reload_submodules(locals())
    del __reload_submodules

from . import _unique_name
from . import _wrapped_text
from . import _developer_extras
from . import _progress
from . import _icons_cache
from . import _preset

from . _unique_name import *
from . _wrapped_text import *
from . _developer_extras import *
from . _progress import *
from . _icons_cache import *
from . _preset import *


__all__ = (
    # file://./_unique_name.py
    "eval_unique_name",

    # file://./_wrapped_text.py
    "eval_text_pixel_dimensions",
    "draw_wrapped_text",

    # file://./_developer_extras.py
    "developer_extras_poll",
    "template_developer_extras_warning",

    # file://./_progress.py
    "progress",

    # file://./_icons_cache.py
    "IconsCache",

    # file://./_preset.py
    "copy_default_presets_from",
    "template_preset",
)
